@@ -13,13 +13,20 @@
 
 struct ena_stats {
 	char name[ETH_GSTRING_LEN];
-	int stat_offset;
+	u16 stat_offset;
+	bool is_atomic;
+	u8 reserved;
 };
 
 struct ena_hw_metrics {
 	char name[ETH_GSTRING_LEN];
 };
 
+enum ena_stat_type {
+	ENA_TX_STAT,
+	ENA_RX_STAT,
+};
+
 #define ENA_STAT_ENA_COM_ADMIN_ENTRY(stat) { \
 	.name = #stat, \
 	.stat_offset = offsetof(struct ena_com_stats_admin, stat) / sizeof(u64) \
@@ -64,6 +71,15 @@ struct ena_hw_metrics {
 	.name = #stat \
 }
 
+#define ENA_STAT_ENTRY_ATOMIC(stat, stat_type) { \
+	.name = #stat, \
+	.stat_offset = offsetof(struct ena_stats_##stat_type, stat) / sizeof(u64), \
+	.is_atomic = true \
+}
+
+#define ENA_STAT_TX_ENTRY_ATOMIC(stat) \
+	ENA_STAT_ENTRY_ATOMIC(stat, tx)
+
 static const struct ena_stats ena_stats_global_strings[] = {
 	ENA_STAT_GLOBAL_ENTRY(total_resets),
 	ENA_STAT_GLOBAL_ENTRY(reset_fail),
@@ -118,7 +134,8 @@ static const struct ena_stats ena_srd_info_strings[] = {
 	ENA_STAT_ENA_SRD_ENTRY(ena_srd_resource_utilization)
 };
 
-static const struct ena_stats ena_stats_tx_strings[] = {
+/* These stats will be printed as accumulated values on reset */
+static const struct ena_stats ena_accum_stats_tx_strings[] = {
 	ENA_STAT_TX_ENTRY(cnt),
 	ENA_STAT_TX_ENTRY(bytes),
 	ENA_STAT_TX_ENTRY(queue_stop),
@@ -127,19 +144,29 @@ static const struct ena_stats ena_stats_tx_strings[] = {
 	ENA_STAT_TX_ENTRY(linearize),
 	ENA_STAT_TX_ENTRY(linearize_failed),
 	ENA_STAT_TX_ENTRY(napi_comp),
-	ENA_STAT_TX_ENTRY(tx_poll),
-	ENA_STAT_TX_ENTRY(doorbells),
 	ENA_STAT_TX_ENTRY(prepare_ctx_err),
 	ENA_STAT_TX_ENTRY(bad_req_id),
 	ENA_STAT_TX_ENTRY(llq_buffer_copy),
 	ENA_STAT_TX_ENTRY(missed_tx),
-	ENA_STAT_TX_ENTRY(unmask_interrupt),
+	ENA_STAT_TX_ENTRY_ATOMIC(pending_timedout_pkts),
 #ifdef ENA_AF_XDP_SUPPORT
 	ENA_STAT_TX_ENTRY(xsk_cnt),
 	ENA_STAT_TX_ENTRY(xsk_bytes),
 	ENA_STAT_TX_ENTRY(xsk_need_wakeup_set),
 	ENA_STAT_TX_ENTRY(xsk_wakeup_request),
 #endif /* ENA_AF_XDP_SUPPORT */
+#ifdef ENA_XDP_MB_SUPPORT
+	ENA_STAT_TX_ENTRY(xdp_frags_exceeded),
+	ENA_STAT_TX_ENTRY(xdp_short_linear_part),
+#endif /* ENA_XDP_MB_SUPPORT */
+};
+
+/* These stats will be printed per queue on reset */
+static const struct ena_stats ena_per_q_stats_tx_strings[] = {
+	ENA_STAT_TX_ENTRY(tx_poll),
+	ENA_STAT_TX_ENTRY(doorbells),
+	ENA_STAT_TX_ENTRY(unmask_interrupt),
+	ENA_STAT_TX_ENTRY(lost_interrupt),
 };
 
 static const struct ena_stats ena_stats_rx_strings[] = {
@@ -169,9 +196,26 @@ static const struct ena_stats ena_stats_rx_strings[] = {
 	ENA_STAT_RX_ENTRY(xdp_invalid),
 	ENA_STAT_RX_ENTRY(xdp_redirect),
 #endif
+#ifdef ENA_LPC_SUPPORT
 	ENA_STAT_RX_ENTRY(lpc_warm_up),
 	ENA_STAT_RX_ENTRY(lpc_full),
 	ENA_STAT_RX_ENTRY(lpc_wrong_numa),
+#endif /* ENA_LPC_SUPPORT */
+#ifdef ENA_PAGE_POOL_SUPPORT
+#ifdef CONFIG_PAGE_POOL_STATS
+	ENA_STAT_RX_ENTRY(pp_alloc_fast),
+	ENA_STAT_RX_ENTRY(pp_alloc_slow),
+	ENA_STAT_RX_ENTRY(pp_alloc_slow_hi_ord),
+	ENA_STAT_RX_ENTRY(pp_alloc_empty),
+	ENA_STAT_RX_ENTRY(pp_alloc_refill),
+	ENA_STAT_RX_ENTRY(pp_alloc_waive),
+	ENA_STAT_RX_ENTRY(pp_cached),
+	ENA_STAT_RX_ENTRY(pp_cache_full),
+	ENA_STAT_RX_ENTRY(pp_ring),
+	ENA_STAT_RX_ENTRY(pp_ring_full),
+	ENA_STAT_RX_ENTRY(pp_released_ref),
+#endif /* CONFIG_PAGE_POOL_STATS */
+#endif /* ENA_PAGE_POOL_SUPPORT */
 #ifdef ENA_AF_XDP_SUPPORT
 	ENA_STAT_RX_ENTRY(xsk_need_wakeup_set),
 	ENA_STAT_RX_ENTRY(zc_queue_pkt_copy),
@@ -190,11 +234,14 @@ static const struct ena_stats ena_stats_ena_com_phc_strings[] = {
 	ENA_STAT_ENA_COM_PHC_ENTRY(phc_cnt),
 	ENA_STAT_ENA_COM_PHC_ENTRY(phc_exp),
 	ENA_STAT_ENA_COM_PHC_ENTRY(phc_skp),
-	ENA_STAT_ENA_COM_PHC_ENTRY(phc_err),
+	ENA_STAT_ENA_COM_PHC_ENTRY(phc_err_dv),
+	ENA_STAT_ENA_COM_PHC_ENTRY(phc_err_ts),
+	ENA_STAT_ENA_COM_PHC_ENTRY(phc_err_eb),
 };
 
 #define ENA_STATS_ARRAY_GLOBAL		ARRAY_SIZE(ena_stats_global_strings)
-#define ENA_STATS_ARRAY_TX		ARRAY_SIZE(ena_stats_tx_strings)
+#define ENA_ACCUM_STATS_ARRAY_TX	ARRAY_SIZE(ena_accum_stats_tx_strings)
+#define ENA_PER_Q_STATS_ARRAY_TX	ARRAY_SIZE(ena_per_q_stats_tx_strings)
 #define ENA_STATS_ARRAY_RX		ARRAY_SIZE(ena_stats_rx_strings)
 #define ENA_STATS_ARRAY_ENA_COM_ADMIN	ARRAY_SIZE(ena_stats_ena_com_admin_strings)
 #define ENA_STATS_ARRAY_ENA_COM_PHC	ARRAY_SIZE(ena_stats_ena_com_phc_strings)
@@ -202,6 +249,10 @@ static const struct ena_stats ena_stats_ena_com_phc_strings[] = {
 #define ENA_STATS_ARRAY_ENA_SRD		ARRAY_SIZE(ena_srd_info_strings)
 #define ENA_METRICS_ARRAY_ENI		ARRAY_SIZE(ena_hw_stats_strings)
 
+/* Used to report number of active and XDP TX queues */
+#define ENA_QUEUE_SIZE_STATS_NUM	2
+
+#ifdef ENA_LPC_SUPPORT /* LPC is the only supported priv-flag */
 static const char ena_priv_flags_strings[][ETH_GSTRING_LEN] = {
 #define ENA_PRIV_FLAGS_LPC	BIT(0)
 	"local_page_cache",
@@ -209,6 +260,7 @@ static const char ena_priv_flags_strings[][ETH_GSTRING_LEN] = {
 
 #define ENA_PRIV_FLAGS_NR ARRAY_SIZE(ena_priv_flags_strings)
 
+#endif /* ENA_LPC_SUPPORT */
 static void ena_safe_update_stat(u64 *src, u64 *dst,
 				 struct u64_stats_sync *syncp)
 {
@@ -220,6 +272,20 @@ static void ena_safe_update_stat(u64 *src, u64 *dst,
 	} while (ena_u64_stats_fetch_retry(syncp, start));
 }
 
+static void ena_safe_accumulate_stat(u64 *src, u64 *dst,
+				     struct u64_stats_sync *syncp)
+{
+	unsigned int start;
+	u64 val;
+
+	do {
+		start = ena_u64_stats_fetch_begin(syncp);
+		val = *src;
+	} while (ena_u64_stats_fetch_retry(syncp, start));
+
+	*(dst) += val;
+}
+
 static void ena_metrics_stats(struct ena_adapter *adapter, u64 **data)
 {
 	struct ena_com_dev *dev = adapter->ena_dev;
@@ -239,71 +305,186 @@ static void ena_metrics_stats(struct ena_adapter *adapter, u64 **data)
 		(*data) += supported_metrics_count;
 
 	} else if (ena_com_get_cap(dev, ENA_ADMIN_ENI_STATS)) {
-		ena_com_get_eni_stats(dev, &adapter->eni_stats);
+		struct ena_admin_eni_stats eni_stats = {};
+
+		ena_com_get_eni_stats(dev, &eni_stats);
 		/* Updating regardless of rc - once we told ethtool how many stats we have
 		 * it will print that much stats. We can't leave holes in the stats
 		 */
 		for (i = 0; i < ENA_STATS_ARRAY_ENI; i++) {
 			ena_stats = &ena_stats_eni_strings[i];
-
-			ptr = (u64 *)&adapter->eni_stats +
-				ena_stats->stat_offset;
-
-			ena_safe_update_stat(ptr, (*data)++, &adapter->syncp);
+			ptr = (u64 *)&eni_stats + ena_stats->stat_offset;
+			**data = *ptr;
+			(*data)++;
 		}
 	}
 
 	if (ena_com_get_cap(dev, ENA_ADMIN_ENA_SRD_INFO)) {
-		ena_com_get_ena_srd_info(dev, &adapter->ena_srd_info);
+		struct ena_admin_ena_srd_info ena_srd_info = {};
+
+		ena_com_get_ena_srd_info(dev, &ena_srd_info);
 		/* Get ENA SRD mode */
-		ptr = (u64 *)&adapter->ena_srd_info;
-		ena_safe_update_stat(ptr, (*data)++, &adapter->syncp);
+		ena_stats = &ena_srd_info_strings[0];
+		ptr = (u64 *)&ena_srd_info + ena_stats->stat_offset;
+		**data = *ptr;
+		(*data)++;
 		for (i = 1; i < ENA_STATS_ARRAY_ENA_SRD; i++) {
 			ena_stats = &ena_srd_info_strings[i];
 			/* Wrapped within an outer struct - need to accommodate an
 			 * additional offset of the ENA SRD mode that was already processed
 			 */
-			ptr = (u64 *)&adapter->ena_srd_info +
-				ena_stats->stat_offset + 1;
-
-			ena_safe_update_stat(ptr, (*data)++, &adapter->syncp);
+			ptr = (u64 *)&ena_srd_info + ena_stats->stat_offset + 1;
+			**data = *ptr;
+			(*data)++;
 		}
 	}
 }
 
-static void ena_queue_stats(struct ena_adapter *adapter, u64 **data)
+#ifdef ENA_PAGE_POOL_SUPPORT
+static void ena_fetch_page_pool_stats(struct ena_adapter *adapter)
+{
+#ifdef CONFIG_PAGE_POOL_STATS
+	struct page_pool_stats stats;
+	struct ena_ring *rx_ring;
+	struct page_pool *pool;
+	int i;
+
+	for (i = 0; i < adapter->max_num_io_queues; i++) {
+		rx_ring = &adapter->rx_ring[i];
+		memset(&stats, 0, sizeof(stats));
+		pool = rx_ring->page_pool;
+
+		if (!pool || !page_pool_get_stats(pool, &stats))
+			continue;
+
+		u64_stats_update_begin(&rx_ring->syncp);
+		rx_ring->rx_stats.pp_alloc_fast = stats.alloc_stats.fast;
+		rx_ring->rx_stats.pp_alloc_slow = stats.alloc_stats.slow;
+		rx_ring->rx_stats.pp_alloc_slow_hi_ord =
+			stats.alloc_stats.slow_high_order;
+		rx_ring->rx_stats.pp_alloc_empty = stats.alloc_stats.empty;
+		rx_ring->rx_stats.pp_alloc_refill = stats.alloc_stats.refill;
+		rx_ring->rx_stats.pp_alloc_waive = stats.alloc_stats.waive;
+
+		rx_ring->rx_stats.pp_cached = stats.recycle_stats.cached;
+		rx_ring->rx_stats.pp_cache_full =
+			stats.recycle_stats.cache_full;
+		rx_ring->rx_stats.pp_ring = stats.recycle_stats.ring;
+		rx_ring->rx_stats.pp_ring_full = stats.recycle_stats.ring_full;
+		rx_ring->rx_stats.pp_released_ref =
+			stats.recycle_stats.released_refcnt;
+		u64_stats_update_end(&rx_ring->syncp);
+	}
+#endif /* CONFIG_PAGE_POOL_STATS */
+}
+
+#endif /* ENA_PAGE_POOL_SUPPORT */
+static void ena_dump_stats_for_queue(struct ena_ring *ring, int stats_count,
+				     u64 **data, enum ena_stat_type stat_type,
+				     const struct ena_stats *stats_strings)
+{
+	const struct ena_stats *ena_stats;
+	u64 *ptr;
+	int i;
+
+	for (i = 0; i < stats_count; i++) {
+		ena_stats = &stats_strings[i];
+
+		if (stat_type == ENA_TX_STAT)
+			ptr = (u64 *)&ring->tx_stats + ena_stats->stat_offset;
+		else
+			ptr = (u64 *)&ring->rx_stats + ena_stats->stat_offset;
+
+		if (ena_stats->is_atomic) {
+			**data = atomic64_read((atomic64_t *)ptr);
+			(*data)++;
+		} else {
+			ena_safe_update_stat(ptr, (*data)++, &ring->syncp);
+		}
+	}
+}
+
+static void ena_accumulate_queues_for_stat(struct ena_adapter *adapter,
+					   const struct ena_stats *ena_stats,
+					   u64 **data,
+					   enum ena_stat_type stat_type)
+{
+	u64 *ptr, accumulated_sum = 0;
+	struct ena_ring *ring;
+	int i;
+
+	for (i = 0; i < adapter->max_num_io_queues; i++) {
+		if (stat_type == ENA_TX_STAT) {
+			ring = &adapter->tx_ring[i];
+			ptr = (u64 *)&ring->tx_stats + ena_stats->stat_offset;
+		} else {
+			ring = &adapter->rx_ring[i];
+			ptr = (u64 *)&ring->rx_stats + ena_stats->stat_offset;
+		}
+
+		if (ena_stats->is_atomic)
+			accumulated_sum += atomic64_read((atomic64_t *)ptr);
+		else
+			ena_safe_accumulate_stat(ptr, &accumulated_sum,
+						 &ring->syncp);
+	}
+
+	ena_safe_update_stat(&accumulated_sum, (*data)++,
+			     &ring->syncp);
+}
+
+static void ena_get_queue_stats(struct ena_adapter *adapter, u64 *data,
+				bool print_accumulated_queue_stats)
 {
 	const struct ena_stats *ena_stats;
 	struct ena_ring *ring;
+	int i;
 
-	u64 *ptr;
-	int i, j;
+	/* Report number of active and XDP TX queues */
+	*data++ = adapter->num_io_queues;
+	*data++ = adapter->xdp_num_queues;
 
-	for (i = 0; i < adapter->num_io_queues + adapter->xdp_num_queues; i++) {
+#ifdef ENA_PAGE_POOL_SUPPORT
+	ena_fetch_page_pool_stats(adapter);
+
+#endif /* ENA_PAGE_POOL_SUPPORT */
+	if (print_accumulated_queue_stats) {
+		for (i = 0; i < ENA_ACCUM_STATS_ARRAY_TX; i++) {
+			ena_stats = &ena_accum_stats_tx_strings[i];
+
+			ena_accumulate_queues_for_stat(adapter, ena_stats,
+						       &data, ENA_TX_STAT);
+		}
+
+		for (i = 0; i < ENA_STATS_ARRAY_RX; i++) {
+			ena_stats = &ena_stats_rx_strings[i];
+
+			ena_accumulate_queues_for_stat(adapter, ena_stats,
+						       &data, ENA_RX_STAT);
+		}
+	}
+
+	for (i = 0; i < adapter->max_num_io_queues; i++) {
 		/* Tx stats */
 		ring = &adapter->tx_ring[i];
 
-		for (j = 0; j < ENA_STATS_ARRAY_TX; j++) {
-			ena_stats = &ena_stats_tx_strings[j];
+		ena_dump_stats_for_queue(ring,
+					 ENA_PER_Q_STATS_ARRAY_TX,
+					 &data, ENA_TX_STAT,
+					 ena_per_q_stats_tx_strings);
 
-			ptr = (u64 *)&ring->tx_stats + ena_stats->stat_offset;
+		ena_dump_stats_for_queue(ring,
+					 ENA_ACCUM_STATS_ARRAY_TX,
+					 &data, ENA_TX_STAT,
+					 ena_accum_stats_tx_strings);
 
-			ena_safe_update_stat(ptr, (*data)++, &ring->syncp);
-		}
-		/* XDP TX queues don't have a RX queue counterpart */
-		if (!ENA_IS_XDP_INDEX(adapter, i)) {
-			/* Rx stats */
-			ring = &adapter->rx_ring[i];
+		/* Rx stats */
+		ring = &adapter->rx_ring[i];
 
-			for (j = 0; j < ENA_STATS_ARRAY_RX; j++) {
-				ena_stats = &ena_stats_rx_strings[j];
-
-				ptr = (u64 *)&ring->rx_stats +
-					ena_stats->stat_offset;
-
-				ena_safe_update_stat(ptr, (*data)++, &ring->syncp);
-			}
-		}
+		ena_dump_stats_for_queue(ring,
+					 ENA_STATS_ARRAY_RX,
+					 &data, ENA_RX_STAT,
+					 ena_stats_rx_strings);
 	}
 }
 
@@ -336,9 +517,9 @@ static void ena_get_phc_stats(struct ena_adapter *adapter, u64 **data)
 	}
 }
 
-static void ena_get_stats(struct ena_adapter *adapter,
-			  u64 *data,
-			  bool hw_stats_needed)
+static u64 *ena_get_base_stats(struct ena_adapter *adapter,
+			       u64 *data,
+			       bool hw_stats_needed)
 {
 	const struct ena_stats *ena_stats;
 	u64 *ptr;
@@ -355,11 +536,12 @@ static void ena_get_stats(struct ena_adapter *adapter,
 	if (hw_stats_needed)
 		ena_metrics_stats(adapter, &data);
 
-	ena_queue_stats(adapter, &data);
 	ena_get_admin_queue_stats(adapter, &data);
 
 	if (ena_phc_is_active(adapter))
 		ena_get_phc_stats(adapter, &data);
+
+	return data;
 }
 
 static void ena_get_ethtool_stats(struct net_device *netdev,
@@ -368,40 +550,77 @@ static void ena_get_ethtool_stats(struct net_device *netdev,
 {
 	struct ena_adapter *adapter = netdev_priv(netdev);
 
-	ena_get_stats(adapter, data, true);
+	data = ena_get_base_stats(adapter, data, true);
+	ena_get_queue_stats(adapter, data, false);
 }
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0)
 #ifdef ENA_HAVE_KERNEL_ETHTOOL_TS_INFO
-static int ena_get_ts_info(struct net_device *netdev, struct kernel_ethtool_ts_info *info)
+static int ena_get_ts_info(struct net_device *netdev,
+			   struct kernel_ethtool_ts_info *info)
 #else
-static int ena_get_ts_info(struct net_device *netdev, struct ethtool_ts_info *info)
+static int ena_get_ts_info(struct net_device *netdev,
+			   struct ethtool_ts_info *info)
 #endif /* ENA_HAVE_KERNEL_ETHTOOL_TS_INFO */
 {
 	struct ena_adapter *adapter = netdev_priv(netdev);
 
+#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 12, 0)
 	info->so_timestamping = SOF_TIMESTAMPING_TX_SOFTWARE |
 				SOF_TIMESTAMPING_RX_SOFTWARE |
 				SOF_TIMESTAMPING_SOFTWARE;
+#else
+	info->so_timestamping = SOF_TIMESTAMPING_TX_SOFTWARE;
+#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(6, 12, 0) */
 
 	info->phc_index = ena_phc_get_index(adapter);
 
+	if (ena_com_hw_timestamping_supported(adapter->ena_dev)) {
+		if (adapter->hw_ts_state.hw_tx_supported !=
+		    ENA_ADMIN_HW_TIMESTAMP_TX_SUPPORT_NONE) {
+			info->so_timestamping |= SOF_TIMESTAMPING_TX_HARDWARE |
+						 SOF_TIMESTAMPING_RAW_HARDWARE;
+
+			info->tx_types = BIT(HWTSTAMP_TX_OFF) |
+					 BIT(HWTSTAMP_TX_ON);
+		}
+
+		if (adapter->hw_ts_state.hw_rx_supported !=
+		    ENA_ADMIN_HW_TIMESTAMP_RX_SUPPORT_NONE) {
+			info->so_timestamping |= SOF_TIMESTAMPING_RX_HARDWARE;
+
+			info->rx_filters = BIT(HWTSTAMP_FILTER_NONE) |
+					   BIT(HWTSTAMP_FILTER_ALL);
+		}
+	}
+
 	return 0;
 }
 
-#endif
-static int ena_get_sw_stats_count(struct ena_adapter *adapter)
+static int ena_get_queue_sw_stats_count(struct ena_adapter *adapter,
+					bool count_accumulated_stats)
 {
-	int count = adapter->num_io_queues * (ENA_STATS_ARRAY_TX + ENA_STATS_ARRAY_RX)
-		    + adapter->xdp_num_queues * ENA_STATS_ARRAY_TX
-		    + ENA_STATS_ARRAY_GLOBAL + ENA_STATS_ARRAY_ENA_COM_ADMIN;
+	int count = adapter->max_num_io_queues * (ENA_ACCUM_STATS_ARRAY_TX +
+		    ENA_PER_Q_STATS_ARRAY_TX + ENA_STATS_ARRAY_RX);
 
-	if (ena_phc_is_active(adapter))
-		count += ENA_STATS_ARRAY_ENA_COM_PHC;
+	if (count_accumulated_stats)
+		count += ENA_STATS_ARRAY_RX + ENA_ACCUM_STATS_ARRAY_TX;
+
+	/* Add entries to report number of active queues + XDP TX queues */
+	count += ENA_QUEUE_SIZE_STATS_NUM;
 
 	return count;
 }
 
+static int ena_get_base_sw_stats_count(struct ena_adapter *adapter)
+{
+	if (ena_phc_is_active(adapter))
+		return ENA_STATS_ARRAY_GLOBAL +
+		       ENA_STATS_ARRAY_ENA_COM_ADMIN +
+		       ENA_STATS_ARRAY_ENA_COM_PHC;
+
+	return ENA_STATS_ARRAY_GLOBAL + ENA_STATS_ARRAY_ENA_COM_ADMIN;
+}
+
 static int ena_get_hw_stats_count(struct ena_adapter *adapter)
 {
 	struct ena_com_dev *dev = adapter->ena_dev;
@@ -423,10 +642,13 @@ int ena_get_sset_count(struct net_device *netdev, int sset)
 
 	switch (sset) {
 	case ETH_SS_STATS:
-		return ena_get_sw_stats_count(adapter) +
+		return ena_get_base_sw_stats_count(adapter) +
+		       ena_get_queue_sw_stats_count(adapter, false) +
 		       ena_get_hw_stats_count(adapter);
+#ifdef ENA_LPC_SUPPORT /* LPC is the only supported priv-flag */
 	case ETH_SS_PRIV_FLAGS:
 		return ENA_PRIV_FLAGS_NR;
+#endif /* ENA_LPC_SUPPORT */
 	}
 
 	return -EOPNOTSUPP;
@@ -461,32 +683,55 @@ static void ena_metrics_stats_strings(struct ena_adapter *adapter, u8 **data)
 	}
 }
 
-static void ena_queue_strings(struct ena_adapter *adapter, u8 **data)
+static void ena_get_queue_strings(struct ena_adapter *adapter, u8 *data,
+				  bool print_accumulated_queue_stats)
 {
 	const struct ena_stats *ena_stats;
 	bool is_xdp;
 	int i, j;
 
-	for (i = 0; i < adapter->num_io_queues + adapter->xdp_num_queues; i++) {
+	ethtool_puts(&data, "num_of_active_io_queues");
+	ethtool_puts(&data, "num_of_xdp_tx_queues");
+
+	if (print_accumulated_queue_stats) {
+		for (i = 0; i < ENA_ACCUM_STATS_ARRAY_TX; i++) {
+			ena_stats = &ena_accum_stats_tx_strings[i];
+
+			ethtool_sprintf(&data, "total_tx_%s", ena_stats->name);
+		}
+
+		for (i = 0; i < ENA_STATS_ARRAY_RX; i++) {
+			ena_stats = &ena_stats_rx_strings[i];
+
+			ethtool_sprintf(&data, "total_rx_%s", ena_stats->name);
+		}
+	}
+
+	for (i = 0; i < adapter->max_num_io_queues; i++) {
 		is_xdp = ENA_IS_XDP_INDEX(adapter, i);
 		/* Tx stats */
-		for (j = 0; j < ENA_STATS_ARRAY_TX; j++) {
-			ena_stats = &ena_stats_tx_strings[j];
+		for (j = 0; j < ENA_PER_Q_STATS_ARRAY_TX; j++) {
+			ena_stats = &ena_per_q_stats_tx_strings[j];
 
-			ethtool_sprintf(data,
-					"queue_%u_%s_%s", i,
+			ethtool_sprintf(&data, "queue_%u_%s_%s", i,
 					is_xdp ? "xdp_tx" : "tx",
 					ena_stats->name);
 		}
 
-		/* In XDP there isn't an RX queue counterpart */
-		if (is_xdp)
-			continue;
+		for (j = 0; j < ENA_ACCUM_STATS_ARRAY_TX; j++) {
+			ena_stats = &ena_accum_stats_tx_strings[j];
 
+			ethtool_sprintf(&data, "queue_%u_%s_%s", i,
+					is_xdp ? "xdp_tx" : "tx",
+					ena_stats->name);
+		}
+
+		/* Rx stats */
 		for (j = 0; j < ENA_STATS_ARRAY_RX; j++) {
 			ena_stats = &ena_stats_rx_strings[j];
 
-			ethtool_sprintf(data, "queue_%u_rx_%s", i, ena_stats->name);
+			ethtool_sprintf(&data, "queue_%u_rx_%s", i,
+					ena_stats->name);
 		}
 	}
 }
@@ -515,9 +760,9 @@ static void ena_get_phc_strings(u8 **data)
 	}
 }
 
-static void ena_get_strings(struct ena_adapter *adapter,
-			    u8 *data,
-			    bool hw_stats_needed)
+static u8 *ena_get_base_strings(struct ena_adapter *adapter,
+				 u8 *data,
+				 bool hw_stats_needed)
 {
 	const struct ena_stats *ena_stats;
 	int i;
@@ -530,11 +775,12 @@ static void ena_get_strings(struct ena_adapter *adapter,
 	if (hw_stats_needed)
 		ena_metrics_stats_strings(adapter, &data);
 
-	ena_queue_strings(adapter, &data);
 	ena_get_admin_strings(&data);
 
 	if (ena_phc_is_active(adapter))
 		ena_get_phc_strings(&data);
+
+	return data;
 }
 
 static void ena_get_ethtool_strings(struct net_device *netdev,
@@ -545,11 +791,14 @@ static void ena_get_ethtool_strings(struct net_device *netdev,
 
 	switch (sset) {
 	case ETH_SS_STATS:
-		ena_get_strings(adapter, data, true);
+		data = ena_get_base_strings(adapter, data, true);
+		ena_get_queue_strings(adapter, data, false);
 		break;
+#ifdef ENA_LPC_SUPPORT /* LPC is the only supported priv-flag */
 	case ETH_SS_PRIV_FLAGS:
 		memcpy(data, ena_priv_flags_strings, sizeof(ena_priv_flags_strings));
 		break;
+#endif /* ENA_LPC_SUPPORT */
 	}
 }
 
@@ -654,7 +903,7 @@ static void ena_update_tx_rings_nonadaptive_intr_moderation(struct ena_adapter *
 
 	val = ena_com_get_nonadaptive_moderation_interval_tx(adapter->ena_dev);
 
-	for (i = 0; i < adapter->num_io_queues; i++) {
+	for (i = 0; i < adapter->num_io_queues + adapter->xdp_num_queues; i++) {
 		adapter->tx_ring[i].interrupt_interval_changed |=
 			(adapter->tx_ring[i].interrupt_interval != val);
 		adapter->tx_ring[i].interrupt_interval = val;
@@ -668,7 +917,7 @@ static void ena_update_rx_rings_nonadaptive_intr_moderation(struct ena_adapter *
 
 	val = ena_com_get_nonadaptive_moderation_interval_rx(adapter->ena_dev);
 
-	for (i = 0; i < adapter->num_io_queues; i++) {
+	for (i = 0; i < adapter->num_io_queues + adapter->xdp_num_queues; i++) {
 		adapter->rx_ring[i].interrupt_interval_changed |=
 			(adapter->rx_ring[i].interrupt_interval != val);
 		adapter->rx_ring[i].interrupt_interval = val;
@@ -751,8 +1000,10 @@ static void ena_get_drvinfo(struct net_device *dev,
 	if (ret < 0)
 		netif_dbg(adapter, drv, dev,
 			  "bus info will be truncated, status = %zd\n", ret);
+#ifdef ENA_LPC_SUPPORT /* LPC is the only supported priv-flag */
 
 	info->n_priv_flags = ENA_PRIV_FLAGS_NR;
+#endif /* ENA_LPC_SUPPORT */
 }
 
 static void ena_get_ringparam(struct net_device *netdev,
@@ -769,7 +1020,7 @@ static void ena_get_ringparam(struct net_device *netdev,
 	ring->tx_max_pending = adapter->max_tx_ring_size;
 	ring->rx_max_pending = adapter->max_rx_ring_size;
 #ifdef ENA_LARGE_LLQ_ETHTOOL
-	if (adapter->ena_dev->tx_mem_queue_type == ENA_ADMIN_PLACEMENT_POLICY_DEV) {
+	if (likely(adapter->ena_dev->tx_mem_queue_type == ENA_ADMIN_PLACEMENT_POLICY_DEV)) {
 		bool large_llq_supported = adapter->large_llq_header_supported;
 
 		kernel_ring->tx_push = true;
@@ -784,7 +1035,7 @@ static void ena_get_ringparam(struct net_device *netdev,
 		kernel_ring->tx_push_buf_len = 0;
 	}
 
-#endif
+#endif /* ENA_LARGE_LLQ_ETHTOOL */
 	ring->tx_pending = adapter->tx_ring[0].ring_size;
 	ring->rx_pending = adapter->rx_ring[0].ring_size;
 }
@@ -833,7 +1084,7 @@ static int ena_set_ringparam(struct net_device *netdev,
 		new_tx_push_buf_len = kernel_ring->tx_push_buf_len;
 
 		placement = adapter->ena_dev->tx_mem_queue_type;
-		if (placement == ENA_ADMIN_PLACEMENT_POLICY_HOST)
+		if (unlikely(placement == ENA_ADMIN_PLACEMENT_POLICY_HOST))
 			return -EOPNOTSUPP;
 
 		if (new_tx_push_buf_len != ENA_LLQ_HEADER &&
@@ -907,9 +1158,17 @@ static u16 ena_flow_data_to_flow_hash(u32 hash_fields)
 	return data;
 }
 
+#ifdef ENA_HAVE_ETHTOOL_RXFH_FIELDS
+static int ena_get_rxfh_fields(struct net_device *netdev,
+			       struct ethtool_rxfh_fields *cmd)
+{
+	struct ena_adapter *adapter = netdev_priv(netdev);
+	struct ena_com_dev *ena_dev = adapter->ena_dev;
+#else
 static int ena_get_rss_hash(struct ena_com_dev *ena_dev,
 			    struct ethtool_rxnfc *cmd)
 {
+#endif /* ENA_HAVE_ETHTOOL_RXFH_FIELDS */
 	enum ena_admin_flow_hash_proto proto;
 	u16 hash_fields;
 	int rc;
@@ -958,9 +1217,18 @@ static int ena_get_rss_hash(struct ena_com_dev *ena_dev,
 	return 0;
 }
 
+#ifdef ENA_HAVE_ETHTOOL_RXFH_FIELDS
+static int ena_set_rxfh_fields(struct net_device *netdev,
+			       const struct ethtool_rxfh_fields *cmd,
+			       struct netlink_ext_ack *extack)
+{
+	struct ena_adapter *adapter = netdev_priv(netdev);
+	struct ena_com_dev *ena_dev = adapter->ena_dev;
+#else
 static int ena_set_rss_hash(struct ena_com_dev *ena_dev,
 			    struct ethtool_rxnfc *cmd)
 {
+#endif /* ENA_HAVE_ETHTOOL_RXFH_FIELDS */
 	enum ena_admin_flow_hash_proto proto;
 	u16 hash_fields;
 
@@ -1127,9 +1395,11 @@ static int ena_set_rxnfc(struct net_device *netdev, struct ethtool_rxnfc *info)
 	int rc = 0;
 
 	switch (info->cmd) {
+#ifndef ENA_HAVE_ETHTOOL_RXFH_FIELDS
 	case ETHTOOL_SRXFH:
 		rc = ena_set_rss_hash(adapter->ena_dev, info);
 		break;
+#endif /* ENA_HAVE_ETHTOOL_RXFH_FIELDS */
 	case ETHTOOL_SRXCLSRLINS:
 		rc = ena_set_steering_rule(adapter->ena_dev, info);
 		break;
@@ -1151,7 +1421,7 @@ static int ena_get_steering_rules_cnt(struct ena_com_dev *ena_dev, struct ethtoo
 		return -EOPNOTSUPP;
 
 	info->rule_cnt = ena_dev->flow_steering.active_rules_cnt;
-	info->data = ena_dev->flow_steering.tbl_size;
+	info->data = ena_dev->flow_steering.tbl_size | RX_CLS_LOC_SPECIAL;
 
 	return 0;
 }
@@ -1288,13 +1558,8 @@ static int ena_get_all_steering_rules(struct ena_com_dev *ena_dev, struct ethtoo
 	return 0;
 }
 
-#if LINUX_VERSION_CODE <= KERNEL_VERSION(3, 2, 0)
-static int ena_get_rxnfc(struct net_device *netdev, struct ethtool_rxnfc *info,
-			 void *rules)
-#else
 static int ena_get_rxnfc(struct net_device *netdev, struct ethtool_rxnfc *info,
 			 u32 *rules)
-#endif
 {
 	struct ena_adapter *adapter = netdev_priv(netdev);
 	int rc = 0;
@@ -1304,9 +1569,11 @@ static int ena_get_rxnfc(struct net_device *netdev, struct ethtool_rxnfc *info,
 		info->data = adapter->num_io_queues;
 		rc = 0;
 		break;
+#ifndef ENA_HAVE_ETHTOOL_RXFH_FIELDS
 	case ETHTOOL_GRXFH:
 		rc = ena_get_rss_hash(adapter->ena_dev, info);
 		break;
+#endif /* ENA_HAVE_ETHTOOL_RXFH_FIELDS */
 	case ETHTOOL_GRXCLSRLCNT:
 		rc = ena_get_steering_rules_cnt(adapter->ena_dev, info);
 		break;
@@ -1326,10 +1593,11 @@ static int ena_get_rxnfc(struct net_device *netdev, struct ethtool_rxnfc *info,
 }
 #endif /* ETHTOOL_GRXRINGS */
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
 static u32 ena_get_rxfh_indir_size(struct net_device *netdev)
 {
-	return ENA_RX_RSS_TABLE_SIZE;
+	struct ena_adapter *adapter = netdev_priv(netdev);
+
+	return get_rss_indirection_table_size(adapter);
 }
 
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 16, 0)
@@ -1342,10 +1610,14 @@ static u32 ena_get_rxfh_key_size(struct net_device *netdev)
 static int ena_indirection_table_set(struct ena_adapter *adapter,
 				     const u32 *indir)
 {
+	u32 table_size = get_rss_indirection_table_size(adapter);
 	struct ena_com_dev *ena_dev = adapter->ena_dev;
 	int i, rc;
 
-	for (i = 0; i < ENA_RX_RSS_TABLE_SIZE; i++) {
+	if (table_size == 0)
+		return -EOPNOTSUPP;
+
+	for (i = 0; i < table_size; i++) {
 		rc = ena_com_indirect_table_fill_entry(ena_dev,
 						       i,
 						       ENA_IO_RXQ_IDX(indir[i]));
@@ -1367,9 +1639,13 @@ static int ena_indirection_table_set(struct ena_adapter *adapter,
 
 static int ena_indirection_table_get(struct ena_adapter *adapter, u32 *indir)
 {
+	u32 table_size = get_rss_indirection_table_size(adapter);
 	struct ena_com_dev *ena_dev = adapter->ena_dev;
 	int i, rc;
 
+	if (table_size == 0)
+		return -EOPNOTSUPP;
+
 	if (!indir)
 		return 0;
 
@@ -1381,7 +1657,7 @@ static int ena_indirection_table_get(struct ena_adapter *adapter, u32 *indir)
 	 * for Tx and uneven indices for Rx. We need to convert the Rx
 	 * indices to be consecutive
 	 */
-	for (i = 0; i < ENA_RX_RSS_TABLE_SIZE; i++)
+	for (i = 0; i < table_size; i++)
 		indir[i] = ENA_IO_RXQ_IDX_TO_COMBINED_IDX(indir[i]);
 
 	return rc;
@@ -1475,14 +1751,14 @@ static int ena_get_rxfh(struct net_device *netdev, u32 *indir, u8 *key)
 
 	return rc;
 }
-#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)/* >= 3.16.0 */
+#else /* >= 3.16.0 */
 static int ena_get_rxfh(struct net_device *netdev, u32 *indir)
 {
 	struct ena_adapter *adapter = netdev_priv(netdev);
 
 	return ena_indirection_table_get(adapter, indir);
 }
-#endif /* >= 3.8.0 */
+#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0) */
 
 #ifdef ENA_HAVE_ETHTOOL_RXFH_PARAM
 static int ena_set_rxfh(struct net_device *netdev,
@@ -1545,7 +1821,7 @@ static int ena_set_rxfh(struct net_device *netdev, const u32 *indir,
 
 	return 0;
 }
-#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0) /* Kernel > 3.16 */
+#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(3, 16, 0) */
 static int ena_set_rxfh(struct net_device *netdev, const u32 *indir)
 {
 	struct ena_adapter *adapter = netdev_priv(netdev);
@@ -1556,9 +1832,7 @@ static int ena_set_rxfh(struct net_device *netdev, const u32 *indir)
 
 	return rc;
 }
-#endif /* Kernel >= 3.8 */
-#endif /* ETHTOOL_GRXFH */
-#ifndef HAVE_RHEL6_ETHTOOL_OPS_EXT_STRUCT
+#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(3, 16, 0) */
 
 #ifdef ETHTOOL_SCHANNELS
 static void ena_get_channels(struct net_device *netdev,
@@ -1603,7 +1877,6 @@ static int ena_set_channels(struct net_device *netdev,
 }
 #endif /* ETHTOOL_SCHANNELS */
 
-#endif /* HAVE_RHEL6_ETHTOOL_OPS_EXT_STRUCT */
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 18, 0)
 static int ena_get_tunable(struct net_device *netdev,
 			   const struct ethtool_tunable *tuna, void *data)
@@ -1644,6 +1917,7 @@ static int ena_set_tunable(struct net_device *netdev,
 	return ret;
 }
 #endif /* 3.18.0 */
+#ifdef ENA_LPC_SUPPORT /* LPC is the only supported priv-flag */
 
 static u32 ena_get_priv_flags(struct net_device *netdev)
 {
@@ -1663,6 +1937,7 @@ static int ena_set_priv_flags(struct net_device *netdev, u32 priv_flags)
 	/* LPC is the only supported private flag for now */
 	return ena_set_lpc_state(adapter, !!(priv_flags & ENA_PRIV_FLAGS_LPC));
 }
+#endif /* ENA_LPC_SUPPORT */
 
 static const struct ethtool_ops ena_ethtool_ops = {
 #ifdef ENA_HAVE_ETHTOOL_OPS_SUPPORTED_COALESCE_PARAMS
@@ -1693,32 +1968,32 @@ static const struct ethtool_ops ena_ethtool_ops = {
 	.get_rxnfc		= ena_get_rxnfc,
 	.set_rxnfc		= ena_set_rxnfc,
 #endif /* ETHTOOL_GRXRINGS */
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
 	.get_rxfh_indir_size    = ena_get_rxfh_indir_size,
-#endif /* >= 3.8.0 */
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 16, 0)
 	.get_rxfh_key_size	= ena_get_rxfh_key_size,
 	.get_rxfh		= ena_get_rxfh,
 	.set_rxfh		= ena_set_rxfh,
-#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
+#ifdef ENA_HAVE_ETHTOOL_RXFH_FIELDS
+	.get_rxfh_fields	= ena_get_rxfh_fields,
+	.set_rxfh_fields	= ena_set_rxfh_fields,
+#endif /* ENA_HAVE_ETHTOOL_RXFH_FIELDS */
+#else
 	.get_rxfh_indir		= ena_get_rxfh,
 	.set_rxfh_indir		= ena_set_rxfh,
-#endif /* >= 3.8.0 */
-#ifndef HAVE_RHEL6_ETHTOOL_OPS_EXT_STRUCT
+#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(3, 16, 0) */
 #ifdef ETHTOOL_SCHANNELS
 	.get_channels		= ena_get_channels,
 	.set_channels		= ena_set_channels,
 #endif /* ETHTOOL_SCHANNELS */
-#endif
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 18, 0)
 	.get_tunable		= ena_get_tunable,
 	.set_tunable		= ena_set_tunable,
 #endif
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0)
 	.get_ts_info		= ena_get_ts_info,
-#endif
+#ifdef ENA_LPC_SUPPORT /* LPC is the only supported priv-flag */
 	.get_priv_flags		= ena_get_priv_flags,
 	.set_priv_flags		= ena_set_priv_flags,
+#endif /* ENA_LPC_SUPPORT */
 };
 
 void ena_set_ethtool_ops(struct net_device *netdev)
@@ -1728,56 +2003,109 @@ void ena_set_ethtool_ops(struct net_device *netdev)
 
 static void ena_dump_stats_ex(struct ena_adapter *adapter, u8 *buf)
 {
+	u8 *base_strings_buf = NULL, *queue_strings_buf = NULL;
+	int base_strings_num, queue_strings_num, i, j, k, rc;
+	u64 *base_data_buf = NULL, *queue_data_buf = NULL;
 	struct net_device *netdev = adapter->netdev;
-	int strings_num, i, rc;
-	u8 *strings_buf;
-	u64 *data_buf;
+	bool print_accumulated_stats = false;
 
-	strings_num = ena_get_sw_stats_count(adapter);
-	if (strings_num <= 0) {
-		netif_err(adapter, drv, netdev, "Can't get stats num\n");
-		return;
-	}
+	if (!buf)
+		print_accumulated_stats = true;
 
-	strings_buf = devm_kcalloc(&adapter->pdev->dev,
-				   ETH_GSTRING_LEN, strings_num,
-				   GFP_ATOMIC);
-	if (!strings_buf) {
+	base_strings_num = ena_get_base_sw_stats_count(adapter);
+	queue_strings_num = ena_get_queue_sw_stats_count(adapter,
+							 print_accumulated_stats);
+
+	base_strings_buf = kcalloc(base_strings_num, ETH_GSTRING_LEN, GFP_ATOMIC);
+	if (!base_strings_buf) {
 		netif_err(adapter, drv, netdev,
-			  "Failed to allocate strings_buf\n");
+			"Failed to allocate base_strings_buf\n");
 		return;
 	}
 
-	data_buf = devm_kcalloc(&adapter->pdev->dev,
-				strings_num, sizeof(u64),
-				GFP_ATOMIC);
-	if (!data_buf) {
+	queue_strings_buf = kcalloc(queue_strings_num, ETH_GSTRING_LEN, GFP_ATOMIC);
+	if (!queue_strings_buf) {
 		netif_err(adapter, drv, netdev,
-			  "Failed to allocate data buf\n");
-		devm_kfree(&adapter->pdev->dev, strings_buf);
-		return;
+			"Failed to allocate queue_strings_buf\n");
+		goto free_resources;
 	}
 
-	ena_get_strings(adapter, strings_buf, false);
-	ena_get_stats(adapter, data_buf, false);
+	base_data_buf = kcalloc(base_strings_num, sizeof(u64), GFP_ATOMIC);
+	if (!base_data_buf) {
+		netif_err(adapter, drv, netdev,
+			"Failed to allocate base_data_buf\n");
+		goto free_resources;
+	}
+
+	queue_data_buf = kcalloc(queue_strings_num, sizeof(u64), GFP_ATOMIC);
+	if (!queue_data_buf) {
+		netif_err(adapter, drv, netdev,
+			"Failed to allocate queue_data_buf\n");
+		goto free_resources;
+	}
+
+	ena_get_base_strings(adapter, base_strings_buf, false);
+	ena_get_queue_strings(adapter, queue_strings_buf, print_accumulated_stats);
+
+	ena_get_base_stats(adapter, base_data_buf, false);
+	ena_get_queue_stats(adapter, queue_data_buf, print_accumulated_stats);
 
 	/* If there is a buffer, dump stats, otherwise print them to dmesg */
-	if (buf)
-		for (i = 0; i < strings_num; i++) {
+	if (buf) {
+		for (i = 0; i < base_strings_num; i++) {
 			rc = snprintf(buf, ETH_GSTRING_LEN + sizeof(u64),
 				      "%s %llu\n",
-				      strings_buf + i * ETH_GSTRING_LEN,
-				      data_buf[i]);
+				      base_strings_buf + i * ETH_GSTRING_LEN,
+				      base_data_buf[i]);
 			buf += rc;
 		}
-	else
-		for (i = 0; i < strings_num; i++)
-			netif_err(adapter, drv, netdev, "%s: %llu\n",
-				  strings_buf + i * ETH_GSTRING_LEN,
-				  data_buf[i]);
 
-	devm_kfree(&adapter->pdev->dev, strings_buf);
-	devm_kfree(&adapter->pdev->dev, data_buf);
+		for (i = 0; i < queue_strings_num; i++) {
+			rc = snprintf(buf, ETH_GSTRING_LEN + sizeof(u64),
+				      "%s %llu\n",
+				      queue_strings_buf + i * ETH_GSTRING_LEN,
+				      queue_data_buf[i]);
+			buf += rc;
+		}
+	} else {
+		for (i = 0; i < base_strings_num; i++)
+			netif_err(adapter, drv, netdev, "%s: %llu\n",
+				  base_strings_buf + i * ETH_GSTRING_LEN,
+				  base_data_buf[i]);
+
+		/* Print number of active and XDP TX queues and accumulated
+		 * queue stats
+		 */
+		for (i = 0; i < ENA_ACCUM_STATS_ARRAY_TX + ENA_STATS_ARRAY_RX +
+				ENA_QUEUE_SIZE_STATS_NUM; i++)
+			netif_err(adapter, drv, netdev, "%s: %llu\n",
+				  queue_strings_buf + i * ETH_GSTRING_LEN,
+				  queue_data_buf[i]);
+
+		/* Print per queue stats */
+		for (j = 0; j < adapter->max_num_io_queues; j++) {
+			for (k = 0; k < ENA_PER_Q_STATS_ARRAY_TX; k++, i++)
+				netif_err(adapter, drv, netdev, "%s: %llu\n",
+					  queue_strings_buf + i * ETH_GSTRING_LEN,
+					  queue_data_buf[i]);
+
+			for (k = 0; k < ENA_ACCUM_STATS_ARRAY_TX; k++, i++)
+				netif_dbg(adapter, drv, netdev, "%s: %llu\n",
+					  queue_strings_buf + i * ETH_GSTRING_LEN,
+					  queue_data_buf[i]);
+
+			for (k = 0; k < ENA_STATS_ARRAY_RX; k++, i++)
+				netif_dbg(adapter, drv, netdev, "%s: %llu\n",
+					  queue_strings_buf + i * ETH_GSTRING_LEN,
+					  queue_data_buf[i]);
+		}
+	}
+
+free_resources:
+	kfree(base_strings_buf);
+	kfree(queue_strings_buf);
+	kfree(base_data_buf);
+	kfree(queue_data_buf);
 }
 
 void ena_dump_stats_to_buf(struct ena_adapter *adapter, u8 *buf)
